/******************************************************************************

  CEdge.C

  Stephen Freund
  freunds@cs.stanford.edu

******************************************************************************/

#include "CEdge.h"
#include "CNode.h"
#include "CCommodity.h"
#include "CEdgeQueue.h"
#include "CHeap.h"
#include "Parameter.h"
#include "Utility.h"
#include <math.h>
#include <stdio.h>
#include "CGraph.h"

/*
 *
 */
CEdge::CEdge()  {
  CEdge(NULL, NULL, 0);
}


/*
 * Just set up local parameters.  We cannot create the edge queues 
 * here.
 */
CEdge::CEdge(CNode *aTail, CNode *aHead, double cap)  {
  itsHead = aHead;
  itsTail = aTail;
  itsOriginalCapacity = itsCapacity = cap;
  itsQueues = NULL;
  itsDerivativeHeap = NULL;
}


/*
 * delete only the fields that were actualyl created.
 */
CEdge::~CEdge()  {
  if (itsQueues)
    delete[] itsQueues;
  if (itsDerivativeHeap)
    delete itsDerivativeHeap;
}	


/*
 * The following two functions are helpers used while debugging
 * the edges. They print the contents of an edge in a semi-readable
 * format.
 */
ostream& operator<<(ostream& out,const CEdge& edge)  {
  out << "(" << edge.itsTail -> Id();
  out << " -> " << edge.itsHead -> Id() << "; " ;
  out << edge.itsCapacity;
  out << "-";
  edge.PrintQueues(out);
  out << ") ";
  return out;
}

/*
 * Print each set of queues for the commodities.
 */
void CEdge::PrintQueues(ostream& out) const  {
  if (itsQueues != NULL) {
    for (int i = 0; i < itsNumCommodities; i++)
      out <<  *itsQueues[i];
  }
}


/*
 * Create and initalize the edge queues.  
 * Also, build the heap.
 */
void CEdge::InitQueues(int numCommodities, CCommodity *comm[])  {
  int i;
  itsNumCommodities = numCommodities;
  itsQueues = new (CEdgeQueue*)[itsNumCommodities];
  itsDerivativeHeap = new CHeap(numCommodities);
  
  for (i=0; i< itsNumCommodities; i++)  {
    itsQueues[i] = new CEdgeQueue(comm[i]);
    itsDerivativeHeap -> Insert( itsQueues[i]->DerivativeAcrossEdge(), i );
  }

}	

/*
 * Sets the size of the queue for the given commodity.  pos
 * indicates whether the change occurs at the head or the tail of
 * this edge.  level is the new level.  This notifies the
 * appropriate Node of the change in queue level so it can determine
 * if we need to recompute the derivative in the heap.
 */
void CEdge::SetQueueSize(EdgePos pos, int index, double level)   { 
  double unitSize = GetSigma() * itsQueues[index] -> GetCommodity() -> 
                          OriginalDemand();

  /* if SetCommodityLevel returns TRUE, a sufficient amount has
     been moved across to warrant recomputing the derivative */
  if (itsQueues[index]->SetCommodityLevel(pos, level))  {
    itsDerivativeHeap -> ChangeKey(index, itsQueues[index] ->
				   DerivativeAcrossEdge(unitSize));
    itsQueues[index] -> ResetChanges();
   }

  (pos == head ? itsHead : itsTail) -> 
    ChangeSizeOfQueue(this, index, (int)(level / unitSize));

}

/* 
 *
 */
void CEdge::AdjustQueueSize(EdgePos pos, int index, double deltaAmount)  { 
  SetQueueSize(pos, index, GetQueueSize(pos,index) + deltaAmount);
}


/*
 * Divide all queue heights and flows in half.
 */
void CEdge::HalveQueueSizes(void)  {
  int i;
  for (i=0; i< itsNumCommodities; i++)  {
    itsQueues[i] -> HalveQueues();
    itsQueues[i] -> HalveFlow();
  }
}


/*
 * Divide all queue heights and flows in half.
 */
void CEdge::DivideQueueSizes(double a)  {
  int i;
  for (i=0; i< itsNumCommodities; i++)  {
    itsQueues[i] -> DivideQueues(a);
    itsQueues[i] -> DivideFlow(a);
  }
}

/*
 * Move the given amount of the commodity with id of index
 * from the tail to the head.
 */
void CEdge::MoveAcrossEdge(int index, int amount)  {
  itsQueues[index] -> MoveCommodityAcross(amount);
}


/*
 * Performs the balance operation across the edge.
 * We store the unit sizes and commodities in temporary 
 * arrays while we move them across.  This saves us 
 * the trouble of recomputing these values when we 
 * put them back at the end of the function. 
 * Just iteratively takes the top element from the heap,
 * determines how much can be moved across, and moves it. 
 * This code can be simplified if we do not wish to be 
 * able to use cutoffs and undirected graphs.  In that case,
 * just remove the code inserted for those special operations.
 */
void CEdge::BalanceEdge(void)  {
  int i;
  CEdgeQueue *theQueue;
  int numQueuesUsed = 0;
  double transferAmount;
  double headlevel, taillevel;
  CCommodity *theCommodity;
  int usedQueues[MaxCommodities];
  double usedUnitSizes[MaxCommodities];  
	
#ifdef LOW_CUTOFF
  int graphNumEdges = itsHead -> GetGraph() -> GetNumEdges();
#endif

  /* Iterate until we cannot move anything more. Stop when
     nothing else to move, nothing will help, or no capacity 
     left. */
  while (itsDerivativeHeap -> NumElemsInHeap() > 0 && 
	 itsDerivativeHeap -> GetMaxKey() > 0 && 
	 itsCapacity > 0) {

    /* Get the index of the commodity we want to move. */
    HeapItemT h = itsDerivativeHeap -> ExtractMax();
    int queueNum = h.itsCommodity;
    
    usedQueues[numQueuesUsed] = queueNum;

    /* compute the unit size. */
    theQueue = itsQueues[queueNum];
    theCommodity = theQueue -> GetCommodity();
    usedUnitSizes[numQueuesUsed] = GetSigma() * theCommodity -> 
      OriginalDemand();
    
    taillevel = theQueue -> GetCommodityLevel(tail);
    headlevel = theQueue -> GetCommodityLevel(head);

     /* compute amount to move */
    transferAmount = Min(usedUnitSizes[numQueuesUsed], itsCapacity);

    if (transferAmount != usedUnitSizes[numQueuesUsed])  {
	numQueuesUsed++;
	break;
   }

/* if undirected, we need figure out which way the flow should go.
 * Also, for enforcing a low cutoff on the height of the queues,
 * we need to figure out which queue is being changed and check the level.
 */
#ifdef UNDIRECTED
    /* if its an undirected graph, we */
    if (theQueue -> RealDerivative(usedUnitSizes[numQueuesUsed]) < 0)  {
      transferAmount *= -1;
#ifdef LOW_CUTOFF
      {
	double cutoff = - GetEpsilon() / graphNumEdges * 
	                   theCommodity -> OriginalDemand();
	if (headlevel < cutoff) { 
	  numQueuesUsed++;
	  continue;
	}
      }
#endif
     } else {
#ifdef LOW_CUTOFF
	double cutoff = - GetEpsilon() / graphNumEdges * 
	                   theCommodity -> OriginalDemand();
	if (taillevel < cutoff) { 
	  numQueuesUsed++;
	  continue;
	}
#endif
      }

#else

/* For the directed case, just look at the height of the tail. */
#ifdef LOW_CUTOFF
    {
      double cutoff = - GetEpsilon() / graphNumEdges * 
	theCommodity -> OriginalDemand();
      if (taillevel < cutoff) { 
	numQueuesUsed++;
	continue;
      }
    }
#endif

#endif

/* If undirected, figure out if we are increasing the flow or reducing it */
#ifdef UNDIRECTED
    double flow = GetFlow(queueNum);
    if (flow > 0)  {
      if  (transferAmount > 0) {
	itsCapacity -= transferAmount;  /* increase flow in pos dir */
      } else {
	itsCapacity += transferAmount;  /* move against flow in pos dir */
      }
    } else {
      if (transferAmount < 0) {
	itsCapacity += transferAmount;  /* increase flow in neg dir */
      } else {
	itsCapacity -= transferAmount;  /* move against flow in neg dir */
      }
    }
#else
    itsCapacity -= transferAmount;
#endif

    /* move the amount across. */
    theQueue -> MoveCommodityAcross(transferAmount);  
    itsTail -> ChangeSizeOfQueue(this, theCommodity -> Id(),
          (int)((taillevel -  transferAmount) / usedUnitSizes[numQueuesUsed]));
    itsHead -> ChangeSizeOfQueue(this, theCommodity -> Id(), 
          (int)((headlevel + transferAmount) / usedUnitSizes[numQueuesUsed]));
    
    numQueuesUsed++;
  }			

  /* Reinsert what we have taken out of the queue. */
  for (i=0;i<numQueuesUsed;i++)  {
    itsQueues[usedQueues[i]] -> ResetChanges();
    itsDerivativeHeap -> Insert(
	itsQueues[usedQueues[i]] -> DerivativeAcrossEdge(usedUnitSizes[i]), 
	usedQueues[i]);
  }
}

/*
 * Print the flows across the edge.
 */
void CEdge::PrintFlow(void)  {
  int i;
  double total = 0;
  printf(" %2d -> %-2d:\n", itsTail->Id(), itsHead->Id());
  for (i=0;i<itsNumCommodities;i++)  {
    if (itsQueues[i] -> GetFlow() != 0)  {
      printf("   %2d:  %-8.4g\n", i, itsQueues[i]->GetFlow());
      total += itsQueues[i] -> GetFlow();
    }
  }
  printf("  total:  %-8.4g\n", total);
  printf("\n");
}



 






