/******************************************************************************

  CEdgeQueue.h

  Stephen Freund
  freunds@cs.stanford.edu

  The CEdgeQueue class maintains a pair of queues across an edge, representing
  the queues for a single commodity at the head and the tail of an edge.

  Create one of these for each commodity on each edge.  When accessing 
  queue levels, the location is specified by the EdgePos enumeration of
  head and tail.  These indicate which queue to use for most of the 
  operations.

  During edge balancing, use MoveCommodityAcrossEdge and DerivativeAcrossEdge
  to perform the necessary operations on a commodity's queues.

  During node balancing, use AdjustCommodityLevel, Get/SetCommodityLevel.

  Each EdgeQueue keeps track of the changes in each queue. These are used
  to determine when the heights of the queues have changed enough to 
  warrant recomputing the derivative for that edge.  Right now, it is set
  to say that the derivative must be computed only when atleast one of
  the queues has changed by atleast sigma * originalDemand.

******************************************************************************/

#ifndef _H_CEdgeQueue
#define _H_CEdgeQueue

#include <iostream.h>
#include "CHeap.h"
#include "CCommodity.h"

class CEdge;
class CCommodity;

typedef enum { head, tail } EdgePos;

class CEdgeQueue  {
  
  private:
    double itsLevel[2];        /* Levels at both ends */
    double change[2];          /* net changes since derivative last computed */
    CCommodity *itsCommodity;  /* the commodity for the queues */
    double itsFlow;            /* flow of commodity across edge */
  
  public:

    /* constructors / destructors */
    CEdgeQueue();
    CEdgeQueue(CCommodity *commodity);
  
    /* setting and getting commodity levels */
    void SetCommodity(CCommodity *comm);
    CCommodity *GetCommodity(void) 	 { return itsCommodity; }
    double GetCommodityLevel(EdgePos pos)  { return itsLevel[pos]; }
    int AdjustCommodityLevel(EdgePos pos, double deltalevel);
    int SetCommodityLevel(EdgePos pos, double level);  
  
    /* monitoring amount by which queues change */
    void ResetChanges(void)               { change[head] = change[tail] = 0; }
    int TimeToChangeKey(void) 
      { return change[head] > itsCommodity -> OriginalDemand() ||
	change[tail] > itsCommodity -> OriginalDemand();  }

    /* for use during balancing across edge */
    int MoveCommodityAcross(double amount);
    double DerivativeAcrossEdge(double transferAdjustment = 0);
    double RealDerivative(double transferAdjustment = 0);

    /* operations at end of phase */
    void HalveFlow(void)         {  itsFlow /= 2; }
    void HalveQueues(void)       { itsLevel[head] /= 2; itsLevel[tail] /= 2; }

    void DivideFlow(double a)     { itsFlow /= a; }
    void DivideQueues(double a)   { itsLevel[head] /= a; itsLevel[tail] /= a; }

    double GetFlow(void)         { return itsFlow; }

    /* I/O */
    friend ostream& operator<<(ostream& out,const CEdgeQueue& q);  
};


#endif

