/******************************************************************************

  CHeap.C

  Stephen Freund
  freunds@cs.stanford.edu

  This implementation of a priority queue use a binary tree reprentation.
  Space for the whole queue is implemented beforehand to avoid the
  over head of memory management.

  Most of this was taken from the big white algorithms book.

******************************************************************************/

#include "CHeap.h"
#include "Utility.h"

/*
 * Assumes a binary heap.
 */
#define Parent(X)		(X >> 1)
#define Left(X)			(X << 1)
#define Right(X)		((X << 1) + 1)


CHeap::CHeap(int maxItems)  {
  itsNumItems = 0;
  itsMaxItems = maxItems;
  itsItems = new HeapItemT[maxItems];
  itsLocationsInHeap = new int[maxItems];
}


void CHeap::Swap(int a, int b)  {
  HeapItemT temp = itsItems[a];
  itsItems[a] = itsItems[b];
  itsLocationsInHeap[itsItems[b].itsCommodity] = a;
  itsItems[b] = temp;
  itsLocationsInHeap[temp.itsCommodity] = b;	
}


void CHeap::Heapify(int node)  {
  int largest, l, r;
  
  while (TRUE) {
    largest = node;
    l = Left(node);
    r = Right(node);
    
    if (l < itsNumItems && itsItems[l].itsKey > itsItems[node].itsKey) 
      largest = l;
    
    if (r < itsNumItems && itsItems[r].itsKey > itsItems[largest].itsKey) 
      largest = r;
    
    if (largest != node)  {
      Swap(node,largest);
      node = largest;
    }
    else 
      return;
  }
} 


HeapItemT CHeap::ExtractMax(void)  {
  HeapItemT returnVal;
  
  if (itsNumItems < 1)
    Error("Heap Underflow");
  
  returnVal = itsItems[0];
  itsItems[0] = itsItems[itsNumItems - 1];
  itsLocationsInHeap[itsItems[itsNumItems - 1].itsCommodity] = -1;
  itsNumItems--;
  Heapify(0);
  return (returnVal);
}	

void CHeap::Insert(double key, int commodity)  {
  int i;
  
  itsNumItems++;
  if (itsNumItems > itsMaxItems)
    Error("Heap overflow");
  
  i = itsNumItems - 1;
  while (i > 0 && itsItems[Parent(i)].itsKey < key) {
    itsItems[i] = itsItems[Parent(i)];
    itsLocationsInHeap[itsItems[Parent(i)].itsCommodity] = i;
    i = Parent(i);
  }
  itsItems[i].itsKey = key;
  itsItems[i].itsCommodity = commodity;
  itsLocationsInHeap[commodity] = i;
}
	
void CHeap::ChangeKey(int commodity, double newKey)  {
  int theItem;
  
  theItem = itsLocationsInHeap[commodity];
  if (theItem == -1)
	return;
  
  if (newKey > itsItems[theItem].itsKey)  {
    while (theItem > 0 && itsItems[Parent(theItem)].itsKey < newKey) {
      itsItems[theItem] = itsItems[Parent(theItem)];
      itsLocationsInHeap[itsItems[Parent(theItem)].itsCommodity] = theItem;
      theItem = Parent(theItem);
    }
    itsItems[theItem].itsCommodity = commodity;
    itsItems[theItem].itsKey = newKey;
    itsLocationsInHeap[commodity] = theItem;
    
  } else  {
    itsItems[theItem].itsKey = newKey;
    Heapify(theItem);
  }
}

ostream& operator<<(ostream& out,const CHeap& heap)  {
  int i;
  out << "size: " << heap.itsNumItems << '\n';
  out << "max:  " << heap.itsMaxItems << '\n';
  out << '\n';
  return(out);
}
		
istream& operator>>(istream& in, CHeap& heap)  {
  int i, numItems;
  
  if (heap.itsItems != NULL)
    delete heap.itsItems;
  
  in >> heap.itsMaxItems;
  
  in >> numItems;
  for (i=0; i < numItems; i++)  {
    double key;
    int commodity;
    in >> key;
    in >> commodity;
    heap.Insert(key, commodity);
  }
	
  return(in);
  
}







