/******************************************************************************
  
  CNode.C

  Stephen Freund
  freunds@cs.stanford.edu

******************************************************************************/

#include "CNode.h"
#include "CEdge.h"
#include "CCommodity.h"
#include "CGraph.h"
#include "Parameter.h"
#include "Utility.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/*
 * Just set the fields of the object
 */
CNodeItem::CNodeItem(CEdge *theEdge, CNodeItem *theNext)  {
  itsEdge = theEdge;
  itsNext = theNext;
}      

ostream& operator<<(ostream& out, const CNodeItem& theNodeItem)  {
  out << theNodeItem.itsEdge;
  out << *(theNodeItem.itsNext);
  return(out);
}




/*
 * Don't create anything yet - wait until all the edges
 * have been added.
 */
CNode::CNode(CGraph *theGraph, int id)  {
  itsGraph = theGraph;
  itsEdges = NULL;
  itsEnteringEdges = NULL;
  itsId = id;
  itsDegree = 0;
  maxValues = NULL;
  minValues = NULL;
}


/*
 * 
 */
CNode::~CNode()  {
  if (itsEdges != NULL) delete itsEdges; 
  if (itsEnteringEdges != NULL) delete itsEnteringEdges;
  if (maxValues != NULL) delete maxValues;
  if (minValues != NULL) delete minValues;
}

/*
 * Just insert it at head of linked list.
 */
void CNode::AddEdge(CNode *dest)  {
  CEdge *newEdge = new CEdge(this, dest);
  CNodeItem *newItem = new CNodeItem(newEdge, itsEdges);
  itsEdges = newItem;
  itsDegree++;
}

/*
 * Same but in other list.
 */
void CNode::AddEnteringEdge(CEdge *incidentEdge)  {
  CNodeItem *newItem = new CNodeItem(incidentEdge, itsEnteringEdges);
  itsEnteringEdges = newItem;
  itsDegree++;
}

/*
 * This is where memory is allocated for the edge queues on each
 * edge, and for the min and max arrays.
 */
void CNode::InitEdgeQueues(int numCommodities, CCommodity *comm[])   {
  CNodeItem *item;
  int i;
	
  /* init edges */
  for (item = itsEdges; item != NULL; item = item -> itsNext)	
    item->itsEdge->InitQueues(numCommodities, comm);
  
  maxValues = new int[numCommodities];
  minValues = new int[numCommodities];

  /* set maxes to extremes they couldn't possibly be */
  for (i=0;i<numCommodities;i++)  {
    maxValues[i] = -1000;
    minValues[i] = 1000;
  }
}

/*
 * NOT USED ANYMORE
 * This just divides it evenly without heed to unit measurements
 */
double CNode::ComputeAmountOfCommodity(int index)  {
  CNodeItem *item;
  double amount = 0;
  double size;
  
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    size = item -> itsEdge -> GetQueueSize(tail, index);
    amount += size;
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)	{
    size = item -> itsEdge -> GetQueueSize(head, index);
    amount += size;
  }
  return(amount);
}



ostream& operator<<(ostream& out,const CNode& node) {
  CNodeItem *item;
  
  out << node.itsId << ": ";
  for (item = node.itsEdges; item != NULL; item = item -> itsNext)	
    out << *( item -> itsEdge);
  
  out << "\n";
  return(out);
}


/*
 * Just see if it is the new max or min value.
 */
void CNode::ChangeSizeOfQueue(CEdge* theEdge, int com, int units)  {
  maxValues[com]  = Max(maxValues[com], units);
  minValues[com]  = Min(minValues[com], units);
}


/*
 * Return the edge that goes to the destination edge,  
 * or NULL if it does not exist.
 */
CEdge *CNode::GetEdge(CNode *dest) const  {
  CNodeItem *item;
  for (item = itsEdges; item != NULL; item = item -> itsNext)	
    if (item->itsEdge->GetHead() == dest)
      break;
  if (item == NULL)
    return(NULL);
  return(item->itsEdge);
}

/*
 * Sum up the heights of all queues for a given commodity, 
 * ignoring any negative values.
 */
double CNode::ComputePositiveAmountOfCommodity(int index)  {
  CNodeItem *item;
  double amount = 0;
  double size;
  
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    size = item -> itsEdge -> GetQueueSize(tail, index);
    if (size > 0)
      amount += size;
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)	{
    size = item -> itsEdge -> GetQueueSize(head, index);
    if (size > 0)
    amount += size;
  }
  return(amount);
}

/* 
 * This just divides it evenly without heed to unit measurements.
 * This is not used anymore.  It is very slow to do all the floating
 * point operations, and doesn't really add anything the speed of
 * the program.
 */
void CNode::DistributeUnitsToEachNode(int commodityIndex, double amount)  {
  CNodeItem *item;
  double unitSize = GetSigma() * (itsGraph -> GetCommodity(commodityIndex) -> 
				  OriginalDemand());
  double average = amount / itsDegree;
  int UnitsToGive = 0;	
  
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    item -> itsEdge -> SetQueueSize(tail, commodityIndex, average);
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)  {
    item -> itsEdge -> SetQueueSize(head, commodityIndex, average);
  }
}

/*
 * Computes the number of sigma * orig_demand units of a commodity are
 * present in all the queues in the node.
 */
int CNode::ComputeAmountOfCommodityByUnits(int index)  {
  CNodeItem *item;
  int amount = 0;
  int size;
  double unitSize = GetSigma() * 
                    (itsGraph -> GetCommodity(index) -> OriginalDemand());
  
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    size =  (int) ((item -> itsEdge -> GetQueueSize(tail, index)) / unitSize);
    amount += size;
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)	{
    size = (int)((item -> itsEdge -> GetQueueSize(head, index)) / unitSize);
    amount += size;
  }
  
  return(amount);
}


/*
 * divide numUnits evenly among the queues in the graph.  First, compute
 * the integral quotient and remainder for amount / degree.  Then go
 * through each queue, setting the value to the average, and possibly one
 * more unit from the remainder. 
 */
void CNode::DistributeUnitsToEachNodeByUnits(int commodityIndex, int numUnits){
  CNodeItem *item;
  double unitSize = GetSigma() * 
          (itsGraph -> GetCommodity(commodityIndex) -> OriginalDemand());
  int averageUnits = (int) floor(numUnits / itsDegree);
  int leftOvers = numUnits - averageUnits * itsDegree;
  int neededUnits;
  int prevUnits;

  /* force remainder to be positive */
  if (leftOvers < 0)  {
    averageUnits -=1;
    leftOvers += itsDegree;
  }
   
  /* reset max and min to impossible values - they will be set to
     correct values below. */
  minValues[commodityIndex] = 1000;
  maxValues[commodityIndex] = -1000;
  

  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    prevUnits = (int)((item -> itsEdge -> 
		       GetQueueSize(tail, commodityIndex)) / unitSize);
    neededUnits = averageUnits - prevUnits;
    if (leftOvers != 0)  {
      leftOvers--;
      neededUnits++;
    }
    item -> itsEdge -> AdjustQueueSize(tail, commodityIndex, 
				       neededUnits * unitSize);
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)	 {
    prevUnits = (int)((item -> itsEdge -> 
		       GetQueueSize(head, commodityIndex)) / unitSize);
    neededUnits = averageUnits - prevUnits;
    if (leftOvers != 0)  {
      leftOvers--;
      neededUnits++;
    }
    item -> itsEdge -> AdjustQueueSize(head, commodityIndex, 
				       neededUnits * unitSize);
  }
}


/*
 * Balancing operation for when the min and max are two units different
 * in height.  First, find all queues with the max and min heights.
 * Then, subtract or add one unit for every pair of these.  If we changed
 * all of either the maxs or the mins, change the maxValue or minValue for
 * that commodity.  Otherwise, there was different numbers of the two 
 * extremes.
 */
void CNode::BalanceFor2Diff(int commodityIndex)  {
  double unitSize = GetSigma() * itsGraph -> GetCommodity(commodityIndex) ->
                    OriginalDemand();
  CEdge* theMaxEdges[MaxDegree];
  EdgePos maxpos[MaxDegree];
  CEdge* theMinEdges[MaxDegree];
  EdgePos minpos[MaxDegree];
  CNodeItem* item;
  int maxVal = (int)maxValues[commodityIndex];
  int minVal = (int)minValues[commodityIndex];
  int maxC = 0;
  int minC = 0;
  int i;
  int rounds;
  int prevUnits;

  for (item = itsEdges; item != NULL; item = item -> itsNext) {
    prevUnits = (int)((item -> itsEdge -> 
		       GetQueueSize(tail, commodityIndex)) / unitSize);
    if (prevUnits == maxVal)  {	
      maxpos[maxC] = tail;
      theMaxEdges[maxC++] = item -> itsEdge;
    }  else
      if (prevUnits == minVal)  {
	minpos[minC] = tail;
	theMinEdges[minC++] = item -> itsEdge;
      } 
  }
  
  for (item = itsEnteringEdges; item != NULL; item = item -> itsNext)	 {
    prevUnits = (int)((item -> itsEdge -> 
		       GetQueueSize(head, commodityIndex)) / unitSize);
    if (prevUnits == maxVal)  {	
      maxpos[maxC] = head;
      theMaxEdges[maxC++] = item -> itsEdge;
    } else
      if (prevUnits == minVal)  {
	minpos[minC] = head;
	theMinEdges[minC++] = item -> itsEdge;
      } 
  }
  
  rounds = Min(maxC, minC);
  if (rounds == maxC)
    maxValues[commodityIndex]--;
  if (rounds == minC)
    minValues[commodityIndex]--;
  
  
  for (i = 0; i < rounds; i++)  {
    theMaxEdges[i] -> AdjustQueueSize(maxpos[i], commodityIndex, -unitSize);
    theMinEdges[i] -> AdjustQueueSize(minpos[i], commodityIndex, unitSize);
  }
}


/*
 * Figure out which balancing method to use, and then do it.
 */
void CNode::BalanceCommodityInNode(int commodityIndex)  {
  int maxVal = maxValues[commodityIndex];
  int minVal = minValues[commodityIndex];
  int amount;  

  if (maxVal - minVal < 2) {
    return;
  }

  if (maxVal - minVal == 2)  {
    BalanceFor2Diff(commodityIndex);
  } else {
    amount = ComputeAmountOfCommodityByUnits(commodityIndex);
    DistributeUnitsToEachNodeByUnits(commodityIndex, amount);
  }
 
}


/*
 * Balance each commodity in the node 
 */
void CNode::BalanceWithinNode(void)  {
  int i;
  
  int numCommodities = itsGraph->GetNumCommodities();
  for (i=0;i<numCommodities; i++)
	BalanceCommodityInNode(i);
}


/*
 * Tell each edge to balance itself
 */
void CNode::BalanceAcrossEdges(void)  {
  CNodeItem *item;
  
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    item -> itsEdge -> BalanceEdge();
  }
}

/*
 * The next four functions find the largest/smallest queue height for
 * a given commodity in either the incoming edges or the outgoing edges.
 * These are used for determining where to place demand changes at 
 * the source and sink.  
 */

CEdge *CNode::FindSmallestQueueForOutEdge(int commodityIndex)  {
  CNodeItem *item;
  CEdge *theEdge;

  if (itsEdges==NULL)
    return NULL;

  theEdge = itsEdges -> itsEdge;
  double size = theEdge -> GetQueueSize(tail, commodityIndex);
  
  for (item = itsEdges -> itsNext; item != NULL; item = item -> itsNext)    {
    double newSize = item -> itsEdge -> GetQueueSize(tail, commodityIndex);
    if (newSize < size) {
      size = newSize;
      theEdge = item -> itsEdge;
    }
  }
  return(theEdge);
}


CEdge *CNode::FindLargestQueueForInEdge(int commodityIndex)  {
  CNodeItem *item;
  CEdge *theEdge;

  if (itsEnteringEdges==NULL)
    return NULL;
  
  theEdge = itsEnteringEdges -> itsEdge;
  double size = theEdge -> GetQueueSize(head, commodityIndex);
  
  for (item = itsEnteringEdges -> itsNext; item != NULL; 
       item = item -> itsNext)	 {
    double newSize = item -> itsEdge -> GetQueueSize(head, commodityIndex);
    if (newSize > size) {
      size = newSize;
      theEdge = item -> itsEdge;
    }
  }
  return(theEdge);
}

CEdge *CNode::FindSmallestQueueForInEdge(int commodityIndex)  {
  CNodeItem *item;
  CEdge *theEdge;

  if (itsEnteringEdges==NULL)
    return NULL;

  theEdge = itsEnteringEdges -> itsEdge;
  double size = theEdge -> GetQueueSize(head, commodityIndex);
  
  for (item = itsEnteringEdges -> itsNext; item != NULL; 
       item = item -> itsNext)	 {
    double newSize = item -> itsEdge -> GetQueueSize(head, commodityIndex);
    if (newSize < size) {
      size = newSize;
      theEdge = item -> itsEdge;
    }
  }
  return(theEdge);
}


CEdge *CNode::FindLargestQueueForOutEdge(int commodityIndex)  {
  CNodeItem *item;
  CEdge *theEdge;

  if (itsEdges==NULL)
    return NULL;

  theEdge = itsEdges -> itsEdge;
  double size = theEdge -> GetQueueSize(tail, commodityIndex);
  
  for (item = itsEdges -> itsNext; item != NULL; item = item -> itsNext) {
    double newSize = item -> itsEdge -> GetQueueSize(tail, commodityIndex);
    if (newSize > size) {
      size = newSize;
      theEdge = item -> itsEdge;
    }
  }
  return(theEdge);
}


/*
 * Increment the capacity on each edge that leaves this node. 
 */
void CNode::AdjustEdgeCapacities(void)  {
  CNodeItem *item;
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    item -> itsEdge -> IncreaseCapacity();
  }
}


/*
 * Cut the capacity in half for each edge that leaves this node.
 */
void CNode::HalveEdgeCapacitiesAndQueues(void)  {
  CNodeItem *item;
  int i;
  CEdge *theEdge;
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    theEdge = item -> itsEdge;
    theEdge -> HalveCapacity();
    theEdge -> HalveQueueSizes();
  }
}

/*
 * Cut the capacity in half for each edge that leaves this node.
 */
void CNode::DivideEdgeCapacitiesAndQueues(double amount)  {
  CNodeItem *item;
  CEdge *theEdge;
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    theEdge = item -> itsEdge;
    theEdge -> DivideCapacity(amount);
    theEdge -> DivideQueueSizes(amount);
  }
}


/*
 * Print the flows for the edges leaving the node.
 */
void CNode::PrintFlows(void)  {
  CNodeItem *item;
  for (item = itsEdges; item != NULL; item = item -> itsNext)	 {
    item->itsEdge -> PrintFlow();
  }
}


  



